#======================================================================================
#== TclNNTPD -- Tcl Coded News Server, v0.92 (alpha)
#======================================================================================
#== Project Name : TclNNTPD
#== Author       : Scott Reilly
#== Organization : Unified Technologies Corporation (http://www.utcorp.net)
#== Conception   : 08 June 1999
#== Last Update  : 15 March 2001
#== Description  : Tcl coded customizable news server
#== Homepage     : http://www.utcorp.net/opensource/tclnntpd/
#== Notes        :
#== 	For RFC977 and working extensions documentation, see 
#==		http://www.ietf.org/rfc.html
#==		http://www.academ.com/academ/nntp/ietf.html
#==		http://www.mibsoftware.com/userkt/nntp/index.html
#==    		http://www.mibsoftware.com/userkt/nntpext/index.html
#==
#== Copyright (c) 1999-2001 by Unified Technologies Corporation
#== All rights reserved.
#== 
#==    This program is free software; you can redistribute it and/or modify
#==    it under the terms of the GNU General Public License as published by
#==    the Free Software Foundation; either version 2 of the License, or
#==    (at your option) any later version.
#==
#==    This program is distributed in the hope that it will be useful,
#==    but WITHOUT ANY WARRANTY; without even the implied warranty of
#==    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#==    GNU General Public License for more details.
#==
#==    You should have received a copy of the GNU General Public License
#==    along with this program; if not, write to the Free Software
#==    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#==
#== This code is licensed under the GNU Public License (GPL)
#==
#== RCS: @(#) $Id: nntpd.tcl,v 0.92 2001/03/15 14:11:23 sreilly Exp $
#======================================================================================

package provide nntpd 0.92

package require Tclx
#-NOTE: check out ::nntpd::init for additional package requires you may need

namespace eval nntpd {
   variable _NNTPD
   
   set _NNTPD(HOST) ""
   set _NNTPD(PORT) "119"
   set _NNTPD(IP) ""
   set _NNTPD(SOCK_IN) "stdin"
   set _NNTPD(SOCK_OUT) "stdout"
   set _NNTPD(MAIL_HOST) ""
   set _NNTPD(NEWS_ROOT) ""
   set _NNTPD(DEBUG) 0
   set _NNTPD(BACKEND) "file"
   set _NNTPD(ENABLE_POST_NOTIFICATION) 0
   set _NNTPD(ENABLE_POST_FORWARDING) 0
   set _NNTPD(ADMIN_EMAIL) ""
   set _NNTPD(DO_AUTH) 0
   set _NNTPD(AUTH_PROC) ""
   set _NNTPD(ALLOW_POSTING) y
   set _NNTPD(DISABLE_DAEMON) 0
   set _NNTPD(ENABLE_MIRRORING) 1
   set _NNTPD(COMPLAIN_EMAIL) ""
   set _NNTPD(MIRROR_PASSWD) ""

   #- Application related variable (not settings)
   set _NNTPD(APPVER) "0.91"
   set _NNTPD(BUFF) ""
   set _NNTPD(GROUP) ""
   set _NNTPD(GROUPCANPOST) "y"
   set _NNTPD(CURRART) 0
   set _NNTPD(ISBUFFERING) 0
   set _NNTPD(BUFFLINES) 0
   #- This is the maximum number of lines for a posted message.  Make it reasonable.  However, we define it so someone
   #-   doesn't just start a post and write enough to fill the system.
   set _NNTPD(MAXBUFFLINES) 4000
   set _NNTPD(have_user) 0
   set _NNTPD(have_pass) 0
   set _NNTPD(input_user) ""
   set _NNTPD(input_pass) ""
   set _NNTPD(MIRRORMODE) 0

   # Export only two commands, the one used to create the nntpd server and the
   #    one to mirror newsgroups
   namespace export nntpd
   namespace export mirror_newsgroups
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::list
#-- Description : Performs the RFC977 command of LIST
#-- Parameters  : a string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::list { LINE } \
{
	set SUBCMD [string tolower [lindex $LINE 1]]
	switch $SUBCMD \
	{
		extensions \
		{
			#::nntpd::send_rsp $SOCK "402 List extensions not supported"
			#-- List extensions MUST be listed singly on a line 
			#--	with a single space at beginning
			::nntpd::send_rsp "202 Extensions supported:"
			::nntpd::send_rsp " ACTIVE"
			::nntpd::send_rsp " LISTGROUP"
			::nntpd::send_rsp " XOVER"
			::nntpd::send_rsp " OVER"
			::nntpd::send_rsp "."
		}
		newsgroups - \
		active - \
		"" \
		{
			::nntpd::send_rsp "215 list of news groups follows"

			foreach PATH [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) *]] \
			{
				set GROUP [file tail $PATH]

				set ALIST [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP \[0-9\]*]]
				if { [llength $ALIST] != 0 } \
				{
					set GROUPLIST [lsort -dictionary -increasing $ALIST]
					set GROUPHI [file tail [lindex $GROUPLIST end]]
					set GROUPLO [file tail [lindex $GROUPLIST 0]]
				} else \
				{
					set GROUPHI 0
					set GROUPLO 0
				}
				if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP .nopost]] } { set GROUPCANPOST "n" } \
				else { set GROUPCANPOST "y" }
				::nntpd::send_rsp "$GROUP $GROUPHI $GROUPLO $GROUPCANPOST"
			}
			::nntpd::send_rsp "."
		}
		default \
		{
			::nntpd::send_rsp "500 Error invoking LIST call - Contact: $::nntpd::_NNTPD(ADMIN_EMAIL)"
		}
	}
	return
#- end proc ::nntpd::list
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::listgroup
#-- Description : Performs the RFC977 extension command of LISTGROUP
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::listgroup { LINE } \
{
	set PARAM [lindex $LINE 1]
	if { $PARAM == "" } \
	{
		if { $::nntpd::_NNTPD(GROUP) == "" } \
		{
			::nntpd::send_rsp "412 Not currently in news group"
		} \
		else \
		{
			set GROUP $::nntpd::_NNTPD(GROUP)
		}
	} \
	else \
	{
		set GROUP $PARAM
	}
	
	if { [file isdirectory [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP]] } \
	{
		set ::nntpd::_NNTPD(GROUP) $GROUP
		if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) .nopost]] } { set ::nntpd::_NNTPD(GROUPCANPOST) "n" } \
		else { set ::nntpd::_NNTPD(GROUPCANPOST) "y" }
		::nntpd::send_rsp "211 list of article numbers follow"
		set ALIST [lsort -dictionary -increasing [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP \[0-9\]*]]]
		foreach mesg $ALIST \
		{
			::nntpd::send_rsp [file tail $mesg]
		}
		::nntpd::send_rsp "."
		set ::nntpd::_NNTPD(CURRART) [lindex $ALIST 1]
	} \
	else \
	{
		::nntpd::send_rsp "411 No such group"
	}
	return
#- end proc ::nntpd::listgroup
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::newgroups
#-- Description : Performs the RFC977 command of NEWGROUPS
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-- Note	: *NOT SERIOUSLY IMPLEMENTED*
#-------------------------------------------------------------------------------------
proc ::nntpd::newgroups { LINE } \
{
	::nntpd::send_rsp "231 list of new newsgroups follow (not implemented):"
	::nntpd::send_rsp "."
#- end proc ::nntpd::newgroups
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::group
#-- Description : Performs the RFC977 command of GROUP
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::group { LINE } \
{
	if { [file isdirectory [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP)]] == 1 } \
	{
		set ::nntpd::_NNTPD(GROUP) [lindex $LINE 1]
		if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) .nopost]] } { set ::nntpd::_NNTPD(GROUPCANPOST) "n" } \
		else { set ::nntpd::_NNTPD(GROUPCANPOST) "y" }
		set ALIST [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) \[0-9\]*]]
		if { [llength $ALIST] != 0 } \
		{
			set ::nntpd::_NNTPD(GROUPLIST) [lsort -dictionary -increasing $ALIST]
			set ::nntpd::_NNTPD(GROUPHI) [file tail [lindex $::nntpd::_NNTPD(GROUPLIST) end]]
			set ::nntpd::_NNTPD(GROUPLO) [file tail [lindex $::nntpd::_NNTPD(GROUPLIST) 0]]
			set ::nntpd::_NNTPD(GROUPLEN) [llength $::nntpd::_NNTPD(GROUPLIST)]
		} else \
		{
			set ::nntpd::_NNTPD(GROUPHI) 0
			set ::nntpd::_NNTPD(GROUPLO) 0
			set ::nntpd::_NNTPD(GROUPLEN) 0
		}
		::nntpd::send_rsp "211 $::nntpd::_NNTPD(GROUPLEN) $::nntpd::_NNTPD(GROUPLO) $::nntpd::_NNTPD(GROUPHI) $::nntpd::_NNTPD(GROUP) group selected"
		set ::nntpd::_NNTPD(CURRART) $::nntpd::_NNTPD(GROUPLO)
	} \
	else \
	{
		::nntpd::send_rsp "411 no such news group"
	}
	return
#- end proc ::nntpd::group
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::next
#-- Description : Performs the RFC977 command of NEXT
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::next { LINE } \
{
	if { $::nntpd::_NNTPD(GROUP) == "" } \
	{
		::nntpd::send_rsp "412 no news group selected"
	} \
	elseif { $::nntpd::_NNTPD(CURRART) == 0 } \
	{
		::nntpd::send_rsp "420 no current article has been selected"
	} \
	else \
	{
		set ALIST [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) \[0-9\]*]]
		if { [llength $ALIST] != 0 } \
		{
			set ::nntpd::_NNTPD(GROUPLIST) [lsort -dictionary -increasing $ALIST]
			set ::nntpd::_NNTPD(GROUPHI) [file tail [lindex $::nntpd::_NNTPD(GROUPLIST) end]]
			set ::nntpd::_NNTPD(GROUPLO) [file tail [lindex $::nntpd::_NNTPD(GROUPLIST) 0]]
			set ::nntpd::_NNTPD(GROUPLEN) [llength $::nntpd::_NNTPD(GROUPLIST)]

			if { $::nntpd::_NNTPD(CURRART) >= $::nntpd::_NNTPD(GROUPLEN) } \
			{
				::nntpd::send_rsp "421 no next article in this group"
			} \
			else \
			{
				incr ::nntpd::_NNTPD(CURRART)
				::nntpd::send_rsp "223 $::nntpd::_NNTPD(CURRART) [::nntpd::getheader Message-ID:] article retrieved"
			}
		} else \
		{
			::nntpd::send_rsp "421 no next article in this group"
		}
	}
	return
#- end proc ::nntpd::next
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::post
#-- Description : Performs the RFC977 command of : POST
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-- Notes	: If you want to disable posting to ALL newsgroups, change the
#--			variable ::nntpd::_NNTPD(ALLOW_POSTING) to 'n'
#--               To disable posting to a particular newsgroup 
#-------------------------------------------------------------------------------------
proc ::nntpd::post { LINE } \
{
	if { $::nntpd::_NNTPD(ALLOW_POSTING) == "n" } \
	{
		::nntpd::send_rsp "440 posting not allowed"
	} \
	elseif { $::nntpd::_NNTPD(GROUPCANPOST) == "n" } \
	{
		#- Posting is disallowed for this newsgroup
		::nntpd::send_rsp "440 posting now allowed"
	} \
	else \
	{
		if { $::nntpd::_NNTPD(ISBUFFERING) == 0 }  \
		{
			#- If not currently buffering, then they have just called POST
			::nntpd::send_rsp "340 send article to be posted.  End with <CR-LF>.<CR-LF>"
			::nntpd::clearBUFF
			set ::nntpd::_NNTPD(ISBUFFERING) 1
			return
		} \
		elseif { $::nntpd::_NNTPD(BUFFLINES) >= $::nntpd::_NNTPD(MAXBUFFLINES) } \
		{
			#- Check to make sure the person isn't posting a message too large.  If so, return error
			::nntpd::clearBUFF
			::nntpd::send_rsp "441 posting failed: Message too large."
			return
		} \
		elseif { [string compare $LINE "."] != 0 } \
		{
			#- Elseif the line isn't a "." then it is part of the posting message
#puts "adding line to buffer: $LINE"
			if { !(($::nntpd::_NNTPD(BUFFLINES) == 0) && ([string trim $LINE] == "")) } \
			{ 
				lappend ::nntpd::_NNTPD(BUFF) $LINE 
				incr ::nntpd::_NNTPD(BUFFLINES)
			}
			return
		}

		#- Otherwise, the message has been fully sent, so continue with posting it
		set ::nntpd::_NNTPD(BUFFLINES) 0
		set ::nntpd::_NNTPD(ISBUFFERING) 0

		#- This section has hack work.  Namely for postings done via the mirroring process.
#		if { [llength $::nntpd::_NNTPD(BUFF)] == 1 } { set ::nntpd::_NNTPD(BUFF) [lindex [lindex $::nntpd::_NNTPD(BUFF) 0] 0] }
#		if { ([llength $::nntpd::_NNTPD(BUFF)] == 2) && ([lindex $::nntpd::_NNTPD(BUFF) 0] == "") } { set ::nntpd::_NNTPD(BUFF) [lindex [lindex $::nntpd::_NNTPD(BUFF) 1] 0] }
#		if { [llength $::nntpd::_NNTPD(BUFF)] == 2 } { set ::nntpd::_NNTPD(BUFF) [lindex [lindex $::nntpd::_NNTPD(BUFF) 0] 0] }
		#- end hack

		#-- If this is a posting to a mirrored newsgroup, see if we are in MIRRORMODE
		set GROUP [::nntpd::getheader Newsgroups:]
		if { ($::nntpd::_NNTPD(MIRRORMODE)) && \
		    ([llength [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) .mirror_*]]] > 0) } \
		{
			set found 0
			foreach grp [split $GROUP ,] \
			{
				if { [string trim $grp] == $::nntpd::_NNTPD(GROUP) } { set found 1; break }
			}
			if { $found == 1 } { set GROUP $::nntpd::_NNTPD(GROUP) }
		}

		if { [string trim $GROUP] == "" } \
		{
#foreach lin $::nntpd::_NNTPD(BUFF) { puts "lin: [string range $lin 0 20]" }
#puts "Here is BUFFER0: [lindex $::nntpd::_NNTPD(BUFF) 0]"
#puts "Here is BUFFER1: [lindex $::nntpd::_NNTPD(BUFF) 1]"
#puts "length=[llength $::nntpd::_NNTPD(BUFF)]"
			::nntpd::send_rsp "441 posting failed - no newsgroup specified in header"
		} \
		elseif { [string first "," $GROUP] != -1 } \
		{
	   		#- The user is requesting posting on multiple newsgroups at once
			::nntpd::send_rsp "500 Cross-posting not allowed.  Post one group at a time."
			
			#- Verify that all newsgroups exist.  If one doesn't, don't post any.
		} \
		elseif { [file isdirectory [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP]] == 0 } \
		{
			::nntpd::send_rsp "500 Newsgroup does not exist"
		} \
		elseif { $::nntpd::_NNTPD(GROUPCANPOST) == "n" } \
		{
			#- Posting is disallowed for this newsgroup
			::nntpd::send_rsp "440 posting now allowed"
		} \
		elseif { ($::nntpd::_NNTPD(MIRRORMODE) == 0) && ([set mirror_file [lindex [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP .mirror_*]] 0]] != "") } \
		{
			#- This newsgroup is a mirror.
			set mirror_file [lindex [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP .mirror_*]] 0]
			set mirrorHost ""
			set mirrorPort "119"
			regexp -- {\.mirror_([^:]+)(:([0-9]+))?} [file tail $mirror_file] match mirrorHost dummy mirrorPort
			if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "This message is to be mirrored to $mirrorHost:$mirrorPort." }
			if { $::nntpd::_NNTPD(ENABLE_MIRRORING) == 0 } \
			{
				#- Fail posting because you have a mirrored newsgroup but mirroring is disabled
				::nntpd::send_rsp "441 posting failed - posting to this mirrored newsgroup is disabled"
			} \
			elseif { $mirrorHost == "" } \
			{
				#- Fail posting because you didn't identify the mirror host
				::nntpd::send_rsp "441 posting failed - mirror host not specified by news server"
			} \
			else \
			{
				#- Connect to mirror host and post the message there
				set err 0
set newART [::nntpd::makeheader]
lappend newART ""
foreach body_item [::nntpd::getbody] { lappend newART $body_item }
set ::nntpd::_NNTPD(BUFF) $newART
				if { [catch {set toMirror [::nntp::nntp $mirrorHost $mirrorPort]} err2] } { set err 1; ::nntpd::send_rsp "441 posting failed - unable to connect to mirror: $err2" } \
				elseif { [catch {$toMirror group $GROUP} err2] } { set err 1; ::nntpd::send_rsp "441 posting failed - mirror host does not have newsgroup" } \
				elseif { [catch {$toMirror post [join $::nntpd::_NNTPD(BUFF) \n]} err2] } { set err 1; ::nntpd::send_rsp "441 posting failed - unable to post to mirror: $err2" }
				catch {$toMirror quit}
				if { $err == 0 } { ::nntpd::send_rsp "240 article received ok" }
			}
		} \
		else \
		{
			set ALIST [lsort -dictionary -increasing [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP \[0-9\]*]]]
			if { $ALIST == "" } { set ALIST [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP 0] }
			set NART [expr [file tail [lindex $ALIST end]] + 1]
			
			set AFILE [open [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP $NART] w]
			
			#- Assemble and send article
			foreach LINE [::nntpd::makeheader $GROUP $NART] { puts $AFILE $LINE }
			puts $AFILE ""
			foreach LINE [::nntpd::getbody] { puts $AFILE $LINE }		
			puts $AFILE "."
			close $AFILE
			
			if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $GROUP $NART]] == 1 } \
			{
				::nntpd::send_rsp "240 article received ok"
			} \
			else \
			{
				::nntpd::send_rsp "441 posting failed"
			}
			
			#-- Check to see if this newsgroup has a cclist file of people to notify for all postings
			if { ($::nntpd::_NNTPD(ENABLE_POST_NOTIFICATION) == 1) && ([set EMAIL_LIST [::nntpd::getNotifyList $GROUP]] != "") } \
			{
				#-- Prepare email text
				set from "News Man <$::nntpd::_NNTPD(ADMIN_EMAIL)>"
				set subject "NEWSGROUP POSTING ALERT : $::nntpd::_NNTPD(HOST) -- $GROUP"
set body "Automatic newsgroup posting notification service --

NEWS HOST 	: $::nntpd::_NNTPD(HOST)
FROM		: [::nntpd::getheader From:]
DATE		: [::nntpd::getheader Date:]
NEWSGROUP	: [::nntpd::getheader Newsgroups:]
LINK		: [::nntpd::make_newsgroup_link]

SUBJECT		: [::nntpd::getheader Subject:]"
					
				#-- Configure email					
				ezsmtp::config -mailhost $::nntpd::_NNTPD(MAIL_HOST)
				
				#-- Send email (one at a time, to protect anonymity) to people in cclist (if not sender)
				foreach ITEM $EMAIL_LIST \
				{
					if {[catch [list ezsmtp::send -from $from -to $ITEM \
						-subject $subject -body $body] err]} {
					}
				}
                        #- end if exists .cclist
			}

			#-- Check to see if this newsgroup has a fwdlist file of people to automatically forward all postings
			if { ($::nntpd::_NNTPD(ENABLE_POST_FORWARDING) == 1) && ([set EMAIL_LIST [::nntpd::getForwardList $GROUP]] != "") } \
			{
				#-- Prepare email text
				set from "News Man <$::nntpd::_NNTPD(ADMIN_EMAIL)>"
				set subject "NEWSGROUP FWD: $GROUP -- [::nntpd::getheader Subject:]"
set body "***NOTICE***
This is a courtesy forward of a newsgroup posting.
Please do not respond to this email.
Contact $::nntpd::_NNTPD(ADMIN_EMAIL) if you no longer wish to receive courtesy
forwards.

***************

NEWS HOST 	: $::nntpd::_NNTPD(HOST)
FROM		: [::nntpd::getheader From:]
DATE		: [::nntpd::getheader Date:]
NEWSGROUP	: [::nntpd::getheader Newsgroups:]
LINK		: [::nntpd::make_newsgroup_link]

SUBJECT		: [::nntpd::getheader Subject:]
					
BODY :

"
foreach ln $::nntpd::_NNTPD(BUFF) { append body "\n$ln" }

				#-- Configure email					
				ezsmtp::config -mailhost $::nntpd::_NNTPD(MAIL_HOST)
				
				#-- Send email (one at a time, to protect anonymity) to people in cclist (if not sender)
				foreach ITEM $EMAIL_LIST \
				{
					if {[catch [list ezsmtp::send -from $from -to $ITEM \
						-subject $subject -body $body] err]} {
					}
				}
                        #- end if exists .fwdlist
			}

		}
	}
	::nntpd::clearBUFF
	return
#- end proc ::nntpd::post
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::getArticle
#-- Description : Performs the RFC977 commands of ARTICLE, HEAD, BODY, STAT
#-- Parameters  : string containing parameters (if applicable),
#--			and the type of the command to perform...
#--			"ART"=ARTICLE,"HEAD"=HEAD,"BODY"=BODY,"STAT"=STAT
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::getArticle { LINE TYPE } \
{
	set PARAM [lindex $LINE 1]
	set ::nntpd::_NNTPD(BUFF) ""
	
	if { $::nntpd::_NNTPD(GROUP) == "" } \
	{
		::nntpd::send_rsp "412 no news group selected"
	} \
	elseif { ($PARAM == "") && ($::nntpd::_NNTPD(CURRART) == 0) } \
	{
		::nntpd::send_rsp "420 no current article has been selected"
	} \
	elseif { ($PARAM == "") || ([regexp {^[0-9]+$} $PARAM] == 1) } \
	{
		#-- We have numeric ID of article, find which to use
		if { $PARAM == "" } \
		{
			set INDEX $::nntpd::_NNTPD(CURRART)
		} \
		else \
		{ 
			set INDEX $PARAM		
		}

		#-- Now determine what it wanted and send it out	
		if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) $INDEX]] == 0 } \
		{		
			::nntpd::send_rsp "423 no such article number in this group"
		} \
		else \
		{ 
			#-- Read in the full article
			set AFILE [open [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) $INDEX] r]
			while { [gets $AFILE ALINE] >= 0 } \
			{
				lappend ::nntpd::_NNTPD(BUFF) $ALINE
			}
			close $AFILE
		
			#-- Send out appropriate response
			set MSGID [::nntpd::getheader Message-ID:]
			if { $MSGID == "" } { set MSGID "<0>" }
			switch $TYPE \
			{
				ART { ::nntpd::send_rsp "220 $INDEX $MSGID article retrieved - head and body follow" }
				HEAD { ::nntpd::send_rsp "221 $INDEX $MSGID article retrieved - head follows" }
				BODY { ::nntpd::send_rsp "222 $INDEX $MSGID article retrieved - body follows" }
				STAT { ::nntpd::send_rsp "223 $INDEX $MSGID article retrieved - request text separately" }
			}

			set ENDHEAD [lsearch $::nntpd::_NNTPD(BUFF) {}]
			set ENDFILE [expr [llength $::nntpd::_NNTPD(BUFF)] - 1]
			if { ($TYPE == "ART") || ($TYPE == "HEAD") } \
			{
				#-- Send out the header
				for { set CT 0 } { $CT < $ENDHEAD } { incr CT } \
				{
					::nntpd::send_rsp [lindex $::nntpd::_NNTPD(BUFF) $CT]
				}
				if { $TYPE == "ART" } \
				{
					::nntpd::send_rsp ""
				}
			}
			if { ($TYPE == "ART") || ($TYPE == "BODY") } \
			{
				#-- Send out the body 				
				for { set NT [expr $ENDHEAD + 1] } { $NT < $ENDFILE } { incr NT } \
				{
					::nntpd::send_rsp [lindex $::nntpd::_NNTPD(BUFF) $NT]
				}				
			}
			if { $TYPE != "STAT" } { ::nntpd::send_rsp "." }

			set ::nntpd::_NNTPD(CURRART) $INDEX
		}
	} \
	elseif { [regexp -nocase -- {^<.*>$} $PARAM] == 1 } \
	{
		#-- We were passed the message-ID of the article they want
		#-- We will have to find the article with this message-id
		set MESGID $PARAM
		set ARTFOUND FALSE
		#-- Search each newsgroup
		set NGLIST [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) *]]
		foreach NWSGRP $NGLIST \
		{
			#-- Search each article
			set ALIST [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $NWSGRP \[0-9\]*]]
			foreach ART $ALIST \
			{
				#-- Read in article to buffer
				set AFILE [open [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) $ART] r]
				while { [gets $AFILE ALINE] >= 0 } \
				{
					lappend ::nntpd::_NNTPD(BUFF) $ALINE
				}
				close $AFILE
				#-- Search each line of only the header part of the article
				set ENDHEAD [lsearch $::nntpd::_NNTPD(BUFF) {}]
				for { set CNT 0 } { $CNT < $ENDHEAD } { incr CNT } \
				{
					#-- If we find the article with the message-id in it
					#-- 	Send out response and the requested text from it
					if { [string first $MESGID $LINE] } \
					{				
						#-- We found the article!
						#-- Send out appropriate response
						switch $TYPE \
						{
							ART { ::nntpd::send_rsp "220 0 $MESGID article retrieved - head and body follow" }
							HEAD { ::nntpd::send_rsp "221 0 $MESGID article retrieved - head follows" }
							BODY { ::nntpd::send_rsp "222 0 $MESGID article retrieved - body follows" }
							STAT { ::nntpd::send_rsp "223 0 $MESGID article retrieved - request text separately" }
						}
			
						set ENDFILE [expr [llength $::nntpd::_NNTPD(BUFF)] - 1]
						if { ($TYPE == "ART") || ($TYPE == "HEAD") } \
						{
							#-- Send out the header
							for { set CT 0 } { $CT < $ENDHEAD } { incr CT } \
							{
								::nntpd::send_rsp [lindex $::nntpd::_NNTPD(BUFF) $CT]
							}
							if { $TYPE == "ART" } \
							{
								::nntpd::send_rsp ""
							}
						}
						if { ($TYPE == "ART") || ($TYPE == "BODY") } \
						{
							#-- Send out the body 				
							for { set NT [expr $ENDHEAD + 1] } { $NT < $ENDFILE } { incr NT } \
							{
								::nntpd::send_rsp [lindex $::nntpd::_NNTPD(BUFF) $NT]
							}				
						}
						if { $TYPE != "STAT" } { ::nntpd::send_rsp "." }
		
						set ARTFOUND TRUE				
						#-- If found, stop searching through lines
						if { $ARTFOUND == "TRUE" } { break }
					}
				}
				#-- If found, stop searching through articles
				if { $ARTFOUND == "TRUE" } { break }
			}
			#-- If found, stop searching through newsgroups
			if { $ARTFOUND == "TRUE" } { break }
		}
		if { $ARTFOUND == "FALSE" } \
		{
			::nntpd::send_rsp "430 no such article found"
		}
	} \
	else \
	{
		::nntpd::send_rsp "502 Service unavailable"
	}
	return
#- end proc ::nntpd::getArticle
}


#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::mode
#-- Description : Performs the RFC977 command of MODE
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::mode { LINE } \
{
	set ::nntpd::_NNTPD(MODE) [lindex LINE 1]
	::nntpd::send_rsp "200 TclNNTPD v$::nntpd::_NNTPD(APPVER) News Server welcomes you! http://www.utcorp.net/opensource/tclnntpd/"
	return
#- end proc ::nntpd::mode
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::help
#-- Description : Performs the RFC977 command of HELP
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::help { LINE } \
{
	::nntpd::send_rsp "100 Help text follows"
	::nntpd::send_rsp "The following is a list of all supported commands:"
	::nntpd::send_rsp "   ARTICLE"
	::nntpd::send_rsp "   AUTHINFO"
	::nntpd::send_rsp "   BODY"
	::nntpd::send_rsp "   DATE"
	::nntpd::send_rsp "   GROUP"
	::nntpd::send_rsp "   HEAD"
	::nntpd::send_rsp "   HELP"
	::nntpd::send_rsp "   IHAVE"
	::nntpd::send_rsp "   LAST"
	::nntpd::send_rsp "   LIST"
	::nntpd::send_rsp "   LISTGROUP"
	::nntpd::send_rsp "   MODE READER"
	::nntpd::send_rsp "   NEWGROUPS (*not fully implemented*)"
	::nntpd::send_rsp "   NEWNEWS (*not fully implemented*)"
	::nntpd::send_rsp "   NEXT"
	::nntpd::send_rsp "   OVER"
	::nntpd::send_rsp "   POST"
	::nntpd::send_rsp "   QUIT"
	::nntpd::send_rsp "   SLAVE"
	::nntpd::send_rsp "   STAT"
	::nntpd::send_rsp "   XOVER"
	::nntpd::send_rsp "For further help try contacting: $::nntpd::_NNTPD(ADMIN_EMAIL)"
	::nntpd::send_rsp "."
	return
#- end proc ::nntpd::help
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::ihave
#-- Description : Performs the RFC977 command of IHAVE
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#--
#-- NOTE	: Since this proc is not widely implemented and therefore not used very
#--		  often, I override it to toggle MIRRORMODE only if IHAVE was called with
#--		  the mirror_passwd specified at startup
#-------------------------------------------------------------------------------------
proc ::nntpd::ihave { LINE } \
{
	if { [lindex $LINE 1] == "$::nntpd::_NNTPD(MIRROR_PASSWD)" } \
	{
		if { $::nntpd::_NNTPD(MIRRORMODE) == 1 } { set ::nntpd::_NNTPD(MIRRORMODE) 0 } \
		else { set ::nntpd::_NNTPD(MIRRORMODE) 1 }
		::nntpd::send_rsp "435 mirror mode enabled"
	} \
	else \
	{
		::nntpd::send_rsp "435 article not wanted - do not send it"
	}
	return
#- end proc ::nntpd::ihave
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::slave
#-- Description : Performs the RFC977 command of SLAVE
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::slave { LINE } \
{
	::nntpd::send_rsp "202 slave status noted"
	return
#- end proc ::nntpd::slave
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::last
#-- Description : Performs the RFC977 command of LAST
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::last { LINE } \
{
	set ::nntpd::_NNTPD(GROUPLO) [file tail [lindex $::nntpd::_NNTPD(GROUPLIST) 0]]
	
	if { $::nntpd::_NNTPD(GROUP) == "" } \
	{
		::nntpd::send_rsp "412 no news group selected"
	} \
	elseif { $::nntpd::_NNTPD(CURRART) == 0 } \
	{
		::nntpd::send_rsp "420 no current article has been selected"
	} \
	elseif { $::nntpd::_NNTPD(CURRART) > $::nntpd::_NNTPD(GROUPLO) } \
	{
		#-Move the current article pointer to the closest prior article
		#-  that exists and that is greater than ::nntpd::_NNTPD(GROUPLO), if possible
		set CHECK [expr $::nntpd::_NNTPD(CURRART) - 1]
		while { ($CHECK >= $::nntpd::_NNTPD(GROUPLO)) && ([file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) $CHECK]] == 0) } \
		{
			incr CHECK -1
		}
		if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) $CHECK]] == 1 } \
		{
			set ::nntpd::_NNTPD(CURRART) $CHECK
			::nntpd::send_rsp "223 $::nntpd::_NNTPD(CURRART) [::nntpd::getheader Message-ID:] article retrieved"	
		} \
		else \
		{
			::nntpd::send_rsp "422 no previous article in this group"
		}
	} \
	else \
	{
		::nntpd::send_rsp "422 no previous article in this group"
	}
	return	
#- end proc ::nntpd::last
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::next
#-- Description : Performs the RFC977 command of NEXT
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-- Notes	: *NOT SERIOUSLY IMPLEMENTED*
#-------------------------------------------------------------------------------------
proc ::nntpd::newnews { LINE } \
{
	::nntpd::send_rsp "230 list of new articles by message-id follow (not implemented)"
	::nntpd::send_rsp "."
	return
#- end proc ::nntpd::newnews
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::date
#-- Description : Performs the RFC977 command of DATE
#-- 			Returns system date and time in form of : YYYYMMDDhhmmss
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::date { LINE } \
{
	::nntpd::send_rsp "111 [clock format [clock seconds] -format %Y%m%d%H%M%S]"
	return
#- end proc ::nntpd::date
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::xover
#-- Description : Performs the RFC977 extension commands of XOVER, OVER
#-- Parameters  : string containing parameters (if applicable)
#-- Returns     : sends response messages and data to ::nntpd::send_rsp
#-------------------------------------------------------------------------------------
proc ::nntpd::xover { LINE } \
{
	set ::nntpd::_NNTPD(BUFF) ""
	set PARAM [lindex $LINE 1]
		
	if { $::nntpd::_NNTPD(GROUP) == "" } \
	{
		::nntpd::send_rsp "412 No news group currently selected"
		return
	}

	set ALIST [lsort -dictionary -increasing [glob -nocomplain [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) \[0-9\]*]]]			
	
	if { $PARAM == "" } \
	{
		if { $::nntpd::_NNTPD(CURRART) == "" } \
		{
			::nntpd::send_rsp "420 No article(s) selected"
			return
		} \
		else \
		{
			set FIRSTART $::nntpd::_NNTPD(CURRART)
			set LASTART [lindex $ALIST end]
		}
	} \
	else \
	{
		#-parse param to determine range of values for mesg search
		set RANGE [split $PARAM -]
		set FIRSTART [lindex $RANGE 0]
		if { [lindex $RANGE 1] == {} } \
		{
			set LASTART [lindex $ALIST end]
		} \
		else \
		{
			set LASTART [lindex $RANGE 1]
		}
	}
	::nntpd::send_rsp "224 list of article numbers follow"

	for { set CT $FIRSTART } { $CT <= $LASTART } { incr CT } \
	{
		set BFILE [open [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) $CT] r]
		while { [gets $BFILE BLINE] >= 0 } \
		{
			lappend ::nntpd::_NNTPD(BUFF) $BLINE
		}
		set subj [::nntpd::getheader Subject:]
		set from [::nntpd::getheader From:]
		set date [::nntpd::getheader Date:]
		set msgid [::nntpd::getheader Message-ID:]
		set refs [::nntpd::getheader References:]
		set size [file size [file join $::nntpd::_NNTPD(NEWS_ROOT) $::nntpd::_NNTPD(GROUP) $CT]]
		set lines [llength $::nntpd::_NNTPD(BUFF)]

		::nntpd::send_rsp "$CT\t$subj\t$from\t$date\t$msgid\t$refs\t$size\t$lines"
		
		set ::nntpd::_NNTPD(BUFF) ""
		close $BFILE
	}
	::nntpd::send_rsp "."
	return
#- end proc ::nntpd::xover
}

#===
#======================================================================================================
#= HELPER PROCS
#======================================================================================================
#===

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::clearBUFF
#-- Description : Clears the article buffer and related variables
#-- Parameters  : n/a
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::clearBUFF {} \
{
	set ::nntpd::_NNTPD(BUFFLINES) 0
	set ::nntpd::_NNTPD(ISBUFFERING) 0
	set ::nntpd::_NNTPD(BUFF) ""
	return
#- end proc ::nntpd::clearBUFF
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::get_cmd
#-- Description : Gets the command name from a string w/ command and its params
#-- Parameters  : a var to store the command in, & the string from 
#--			the socket
#-- Returns     : stores command name in CMD
#-------------------------------------------------------------------------------------
proc ::nntpd::get_cmd { LINE } \
{
#	gets $::nntpd::_NNTPD(SOCK_IN) LINE
	string trim $LINE
	set CMD [string trim [string tolower [lindex $LINE 0]]]
        #- Output messages if debugging
        if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Client : $LINE" }

	return $CMD
#- end ::nntpd::get_cmd
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::getbody
#-- Description : Gets the body of the current article
#-- Parameters  : n/a
#-- Returns     : the body of the current article
#-------------------------------------------------------------------------------------
proc ::nntpd::getbody {} \
{
	return [lrange $::nntpd::_NNTPD(BUFF) [expr [lsearch $::nntpd::_NNTPD(BUFF) ""] + 1] end]
#- end proc ::nntpd::getbody
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::getForwardList
#-- Description : Gets the list of email addresses to be forward a copy of new postings in the specified newsgroup
#-- Parameters  : NEWSGROUP - the newsgroup who's list of forwarding emails you want
#-- Returns     : the body of the current article
#-------------------------------------------------------------------------------------
proc ::nntpd::getForwardList { NEWSGROUP } \
{
	set EMAIL_LIST ""
	#-- Read in the list of email addresses to notify
	if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $NEWSGROUP .fwdlist]] } \
	{
		set CC_FILE [open [file join $::nntpd::_NNTPD(NEWS_ROOT) $NEWSGROUP .fwdlist] r]
		while { [gets $CC_FILE CC_LINE] >= 0 } \
		{
			#-- Don't send email alert to person posting the message
			if { [string match *$CC_LINE* [::nntpd::getheader From:]] == 0 } \
			{
				lappend EMAIL_LIST $CC_LINE
			}
		}
		close $CC_FILE
	}
	return $EMAIL_LIST
#- end proc ::nntpd::getForwardList
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::getheader
#-- Description : Gets the value of the header parameter sent to it from the
#--			current article
#-- Parameters  : optional argument; the exact header value you want extracted from 
#--			the article (e.g. "Message-ID:")
#--		  if not specified, returns the entire header
#-- Returns     : the header information following the header value
#-------------------------------------------------------------------------------------
proc ::nntpd::getheader { {HDR ""} } \
{
	set HDRS ""
	foreach LINE $::nntpd::_NNTPD(BUFF) \
	{
		if { [string trim $LINE] == "" } \
		{
		   if { $HDR == "" } { return $HDRS } \
		   else { return "" }
		}
		lappend HDRS "$LINE"
		if { $HDR != "" } \
		{
		   set HDR [string trimright [string tolower $HDR] :]
		   if { [regexp -- {^([^:]+):(.*)$} $LINE match newHDR GRP] != 0 } \
	 	   {
		      if { [string tolower $newHDR] == $HDR } { return [string trim $GRP] }
		   }
		}
	}
	return ""
#- end proc ::nntpd::getheader
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::getNotifyList
#-- Description : Gets the list of email addresses to be notified for a posting in the specified newsgroup
#-- Parameters  : NEWSGROUP - the newsgroup who's list of notify emails you want
#-- Returns     : the body of the current article
#-------------------------------------------------------------------------------------
proc ::nntpd::getNotifyList { NEWSGROUP } \
{
	set EMAIL_LIST ""
	#-- Read in the list of email addresses to notify
	if { [file exists [file join $::nntpd::_NNTPD(NEWS_ROOT) $NEWSGROUP .cclist]] } \
	{
	   	set CC_FILE [open [file join $::nntpd::_NNTPD(NEWS_ROOT) $NEWSGROUP .cclist] r]
		while { [gets $CC_FILE CC_LINE] >= 0 } \
		{
			#-- Don't send email alert to person posting the message
			if { [string match *$CC_LINE* [::nntpd::getheader From:]] == 0 } \
			{
				lappend EMAIL_LIST $CC_LINE
			}
		}
		close $CC_FILE
	}
	return $EMAIL_LIST
#- end proc ::nntpd::getNotifyList
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::make_newsgroup_link
#-- Returns     : string for newsgroup link in automatic email notification
#-------------------------------------------------------------------------------------
proc ::nntpd::make_newsgroup_link {} \
{
	if { ($::nntpd::_NNTPD(PORT) == 119) || ([string match $::nntpd::_NNTPD(PORT) nntp] == 1) } \
	{
		return "news://$::nntpd::_NNTPD(HOST)/[::nntpd::getheader Newsgroups:]"
	} \
	else \
	{
		return "news://$::nntpd::_NNTPD(HOST):$::nntpd::_NNTPD(PORT)/[::nntpd::getheader Newsgroups:]"
	}
#- end proc ::nntpd::make_newsgroup_link
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::makeheader
#-- Description : Makes a new header for the current article by contributing server added
#--			header values (includes user provided headers)
#-- Parameters  : both parameters are optional.  If one is specified, they both need to be
#-- 			specified.  Needed only for local posting, not for posting to mirror source
#--			GROUP - the name of the local group being posted to
#--			ARTNUM - the article number in the local group the article will be written to
#-- Returns     : the new header
#-------------------------------------------------------------------------------------
proc ::nntpd::makeheader { {GROUP ""} {ARTNUM ""} } \
{
	set newheader ""
	set dont_get_from_client {Path Message-ID Xref X-Trace X-Complaints-To NNTP-Posting-Host NNTP-Posting-Date Lines X-Mirror}
	set dont_get_from_client [string tolower $dont_get_from_client]
	if { [set oldPath [::nntpd::getheader Path:]] == "" } \
	{ 
		set pattern "{^$::nntpd::_NNTPD(HOST)}"
		if { [regexp -- $pattern $oldPath] == 0 } { lappend newheader "Path: $::nntpd::_NNTPD(HOST)" }
	} \
	else { lappend newheader "Path: $::nntpd::_NNTPD(HOST)!$oldPath" }
        set HDRS [::nntpd::getheader]
	#- Pass along all of the client provided headers, except for those the server is supposed to set
	foreach HDR $HDRS \
	{
	   #- Get the header tag name
	   if { [regexp -- {^([^:]+):(.*)$} $HDR match HDRtag HDRval] != 0 } \
	   {
	      #- If it isn't in the exclude list, it can be passed along
	      if { [lsearch $dont_get_from_client [string tolower $HDRtag]] < 0 } \
	      {
	         lappend newheader $HDR
	      }
	   } \
	   else { lappend newheader $HDR }
	}
	lappend newheader "Lines: [llength [::nntpd::getbody]]"
	#- If the client did not provide a Message-ID: (i.e. Outlook Express), generate one
	if { [set oldMsgID [::nntpd::getheader Message-ID:]] == "" } \
	{
	   lappend newheader "Message-ID: <[clock seconds].[pid]@$::nntpd::_NNTPD(HOST)>"
	} \
	else { lappend newheader "Message-ID: $oldMsgID" }
	if { ($GROUP != "") && ($ARTNUM != "") } \
	{ 
		lappend newheader "NNTP-Posting-Host: $::nntpd::_NNTPD(HOST)"
		lappend newheader "NNTP-Posting-Date: [clock format [clock seconds]]"
		lappend newheader "X-Complaints-To: $::nntpd::_NNTPD(COMPLAIN_EMAIL)" 
	}
	if { [set xmirror [::nntpd::getheader X-Mirror:]] != "" } { lappend newheader "X-Mirror: $xmirror" }
	if { ($GROUP != "") && ($ARTNUM != "") } { lappend newheader "Xref: $::nntpd::_NNTPD(HOST) $GROUP:$ARTNUM" }
	return $newheader
#- end proc ::nntpd::makeheader
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::send_rsp
#-- Description : Sends a string through the socket
#-- Parameters  : the string to send through the socket
#-- Returns     : outputs string to the socket
#-------------------------------------------------------------------------------------
proc ::nntpd::send_rsp { RSP } \
{
        #- Output messages if debugging
        if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Server : $RSP" }
	
#fconfigure $::nntpd::_NNTPD(SOCK_OUT) -buffering none -translation crlf
puts $::nntpd::_NNTPD(SOCK_OUT) $RSP
#	puts $SOCK $RSP
#	flush stdout
#- end proc ::nntpd::send_rsp
}

#===
#=================================================================================================================
#= END OF RFC 977 (& Extentions) IMPLEMENTATION CODE
#=================================================================================================================
#===

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::commHandler
#-- Description : Handles communication from client
#-- Parameters  : sock - the socket connection to the client
#--               LINE - the string sent from the client
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::commHandler {sock LINE} \
{
        if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Received message from $sock : $LINE" }

	if { $::nntpd::_NNTPD(ISBUFFERING) } { set CMD xxx } \
	else { set CMD [::nntpd::get_cmd $LINE] }

#-If authentication is required
	if { $::nntpd::_NNTPD(IS_AUTHORIZED) == 0 } \
	{
		if { $CMD == "authinfo" } \
		{
			if { [llength $LINE] == 3 } \
			{
				if { [string tolower [lindex $LINE 1]] == "user" } \
				{
					if { $::nntpd::_NNTPD(have_user) == 1 } \
					{
						::nntpd::send_rsp "482 Authentication rejected -- AUTHINFO user already sent"
						set ::nntpd::_NNTPD(have_user) 0
						set ::nntpd::_NNTPD(input_user) ""
						set ::nntpd::_NNTPD(input_pass) ""
					} \
					else \
					{
						set ::nntpd::_NNTPD(have_user) 1
						set ::nntpd::_NNTPD(input_user) [lindex $LINE 2]
						::nntpd::send_rsp "381 Password requested"
					}
				} \
				elseif { [string tolower [lindex $LINE 1]] == "pass" } \
				{
					if { $::nntpd::_NNTPD(have_user) == 1 } \
					{
						set ::nntpd::_NNTPD(input_pass) [lindex $LINE 2]
						#- Call the authentication procedure.  If error with call, auth is 0
						if { [catch {set auth [eval {$::nntpd::_NNTPD(AUTH_PROC) $::nntpd::_NNTPD(input_user) $::nntpd::_NNTPD(input_pass)}]}] } { set auth 0 }
						if { $auth == 1 } \
						{
							::nntpd::send_rsp "281 Authentication accepted"
							set ::nntpd::_NNTPD(IS_AUTHORIZED) 1
						} \
						else \
						{
							::nntpd::send_rsp "502 Permission denied -- Invalid username and/or password"
							set ::nntpd::_NNTPD(have_user) 0
							set ::nntpd::_NNTPD(input_user) ""
							set ::nntpd::_NNTPD(input_pass) ""
						}
					} \
					else \
					{
						::nntpd::send_rsp "482 Authentication rejected -- AUTHINFO user not supplied"
						set ::nntpd::_NNTPD(have_user) 0
						set ::nntpd::_NNTPD(input_user) ""
						set ::nntpd::_NNTPD(input_pass) ""
					}
				} \
				else \
				{
					::nntpd::send_rsp "500 Invalid AUTHINFO command"
					set ::nntpd::_NNTPD(have_user) 0
					set ::nntpd::_NNTPD(input_user) ""
					set ::nntpd::_NNTPD(input_pass) ""
				}
			} \
			else \
			{
				::nntpd::send_rsp "500 Invalid AUTHINFO command"
				set ::nntpd::_NNTPD(have_user) 0
				set ::nntpd::_NNTPD(input_user) ""
				set ::nntpd::_NNTPD(input_pass) ""	
			}
		} \
		else \
		{
			::nntpd::send_rsp "480 Authorization required"
		}
	} \
	else \
	{
#-If authentication is not required, or was successful
		switch $CMD \
		{
			article { ::nntpd::getArticle $LINE "ART" }
			authinfo \
			{
				::nntpd::send_rsp "482 Authentication attempt rejected -- Already authenticated"
			}
			body { ::nntpd::getArticle $LINE "BODY" }
			date { ::nntpd::date $LINE }
			group { ::nntpd::group $LINE }
			head { ::nntpd::getArticle $LINE "HEAD" }
			help { ::nntpd::help $LINE }
			ihave { ::nntpd::ihave $LINE }
			last { ::nntpd::last $LINE }
			list { ::nntpd::list $LINE }
			listgroup { ::nntpd::listgroup $LINE }
			mode { ::nntpd::mode $LINE }
			newgroups { ::nntpd::newgroups $LINE }
			newnews { ::nntpd::newnews $LINE }
			next { ::nntpd::next $LINE }
			over { ::nntpd::xover $LINE }
			post { ::nntpd::post $LINE }
			quit \
			{
				::nntpd::send_rsp "205 Goodbye!  Closing connection"
				if { [catch {close $sock} result] } { puts "Error closing socket: $result" }
				catch {unset ::nntpd::_NNTPD(addr,$sock)}
				if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Child (pid=[pid]) exiting via QUIT" }
				exit 0
			}
			slave { ::nntpd::slave $LINE }
			stat { ::nntpd::getArticle $LINE "STAT" }
			xover { ::nntpd::xover $LINE }
			"" {}
			default \
			{
				if { $::nntpd::_NNTPD(ISBUFFERING) == 1 } \
				{
				puts "default sending to post"
				   ::nntpd::post "$LINE"
				} \
				else \
				{
				   ::nntpd::send_rsp "500 Invalid command - Contact: $::nntpd::_NNTPD(ADMIN_EMAIL)"
				}
			}
		}
	}
	return
#-end proc ::nntpd::commHandler
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::nntpd
#-- Description : Instantiate an nntpd server
#-- Parameters  : args - the command line arguments
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::nntpd {args} \
{
   #- Initialize the server with command line arguments
   ::nntpd::init "$args"

   #- Daemonize server
   if { $::nntpd::_NNTPD(DISABLE_DAEMON) == 0 } \
   {
      ::nntpd::daemonize
      signal ignore  SIGHUP
      signal unblock {QUIT TERM}
      signal trap    {QUIT TERM} ::nntpd::shutdown
   }

   #- Create a server listening for $::nntpd::_NNTPD(HOST) (if defined) on port $::nntpd::_NNTPD(PORT)
   if { $::nntpd::_NNTPD(HOST) == "" } \
   {
      #- If no HOST was defined by the user, bind the default interface
      set ::nntpd::_NNTPD(SERVER_SOCK) [socket -server accept $::nntpd::_NNTPD(PORT)]
   } \
   else \
   {
      set ::nntpd::_NNTPD(SERVER_SOCK) [socket -server nntpd_accept -myaddr $::nntpd::_NNTPD(HOST) $::nntpd::_NNTPD(PORT)]
   }
   set server_info [fconfigure $::nntpd::_NNTPD(SERVER_SOCK) -sockname]
   set ::nntpd::_NNTPD(IP) [lindex $server_info 0]
   set ::nntpd::_NNTPD(HOST) [lindex $server_info 1]
   if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "TclNNTPD v$::nntpd::_NNTPD(APPVER) started for $::nntpd::_NNTPD(HOST):$::nntpd::_NNTPD(PORT) on [clock format [clock seconds]]" }
   #- Handle events until variable forever is set (which is never)
   vwait forever
#-end proc ::nntpd::nntpd
}

#-------------------------------------------------------------------------------------
#-- PROC        : nntpd_accept
#-- Description : Accepts connections from clients and forks
#-- Parameters  : the argurments are passed by the system upon client connection
#--       sock - the socket connection to the client
#--       addr - the IP address of the client
#--       port - the port of the client
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc nntpd_accept {sock addr port} \
{
   if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Accept $sock from $addr port $port" }
   if { [set cpid [fork]] > 0 } \
   {
      #- Parent executes this code

      #- Call signal handler to when SIGCHLD received
      signal -restart trap {SIGCHLD} ::nntpd::sigchld_handler
#      #- Parent closes its connection to incoming connection
#      close $sock
set ::nntpd::_NNTPD(sock,$cpid) $sock
#puts "Parent just set ::nntpd::_NNTPD(sock,$cpid) to $sock"
      if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Parent (pid=[pid]) returning from forking $cpid" }
      return
   }
   #- Child executes this code
   if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Child (pid=[pid]) resuming fork commands" }
   #- Close the child's connection to the server socket; otherwise it might try to service new connections
   close $::nntpd::_NNTPD(SERVER_SOCK)
#puts "Child: just closed listening socket - $::nntpd::_NNTPD(SERVER_SOCK)"
   #- Child ignores its own SIGCHLD signal
   signal ignore SIGCHLD
   set ::nntpd::_NNTPD(SOCK_IN) $sock
   set ::nntpd::_NNTPD(SOCK_OUT) $sock
   set ::nntpd::_NNTPD(addr,$sock) {$addr $port}
   fconfigure $sock -buffering line -blocking 0
#puts "Child: listening on sock $sock"
   fileevent $sock readable [list ::nntpd::svcHandler $sock]
   ::nntpd::send_rsp "200 TclNNTPD News Server v$::nntpd::_NNTPD(APPVER) welcomes you! http://www.utcorp.net/opensource/tclnntpd/"
#-end proc nntpd_accept
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::svcHandler
#-- Description : Passes on data sent from client and handles unexpected socket errors
#-- Parameters  : sock - the socket connection to the client
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::svcHandler {sock} \
{
   if { [eof $sock] || [catch {gets $sock line}] } \
   {
      close $sock
      puts "Close $::nntpd::_NNTPD(addr,$sock)"
      unset ::nntpd::_NNTPD(addr,$sock)
      if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Child (pid=[pid]) exit via svcHandler" }
      exit 0
   } \
   else \
   {
      ::nntpd::commHandler $sock $line
   }
#-end proc ::nntpd::svcHandler
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::sigchld_handler
#-- Description : Waits SIGCHLD signals, thereby preventing them from becoming zombies
#-- Parameters  : n/a
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::sigchld_handler {} \
{
   set zombie [wait -nohang -untraced]
   if { $::nntpd::_NNTPD(DEBUG) == 1 } { puts "Parent: Cleaned up dead child: ($zombie)" }
   #- Parent closes its connection to the client of the dead child
   close $::nntpd::_NNTPD(sock,[lindex $zombie 0])
#puts "Parent closed: $::nntpd::_NNTPD(sock,[lindex $zombie 0])"
   unset ::nntpd::_NNTPD(sock,[lindex $zombie 0])
   return
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::shutdown
#-- Description : Cleanup code when program is terminated
#-- Parameters  : n/a
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::shutdown {} \
{
   exit 0
   return
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::daemonize
#-- Description : Daemonizes the server
#-- Parameters  : n/a
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::daemonize {} \
{
   #- Close the stdio sockets
   close stdin
   close stdout
   close stderr
   
   #- Parent terminates
   if {[fork]} {exit 0}
   
   #- Become session leader
   id process group set
   
   #- First child terminates
   if {[fork]} {exit 0}
   
   #- Reopen stdio sockets and point to null
   set fd [open /dev/null r]
   set fd [open /dev/null w]
   set fd [open /dev/null w]
   
   #- Change working directory
   cd /

   #- Clear file mode creatoin mask
   umask 022
   return [id process]
#- end proc ::nntpd::daemonize
}

#-------------------------------------------------------------------------------------
#-- PROC        : ::nntpd::init
#-- Description : Parses and interprets command line arguments to initialize server settings
#-- Parameters  : args - command line option/value setting pairs
#-- Returns     : n/a
#-------------------------------------------------------------------------------------
proc ::nntpd::init {args} \
{
   #- Parse command line options
   foreach {opt val} [string trimright [string trimleft $args \{] \}] \
   {
      switch -- $opt \
      {
         "-host" { set ::nntpd::_NNTPD(HOST) $val }
         "-port" { set ::nntpd::_NNTPD(PORT) $val }
         "-mailhost" { set ::nntpd::_NNTPD(MAIL_HOST) $val }
         "-admin_email" { set ::nntpd::_NNTPD(ADMIN_EMAIL) $val }
         "-news_root" { set ::nntpd::_NNTPD(NEWS_ROOT) $val }
         "-debug" { if { [regexp -- {^[1yY]([eE][sS])?$} $val] } { set ::nntpd::_NNTPD(DEBUG) 1 } else { set ::nntpd::_NNTPD(DEBUG) 0 } }
         "-do_auth" { if { [regexp -- {^[1yY]([eE][sS])?$} $val] } { set ::nntpd::_NNTPD(DO_AUTH) 1 } else { set ::nntpd::_NNTPD(DO_AUTH) 0 } }
         "-allow_posting" { if { [regexp -- {^[1yY]([eE][sS])?$} $val] } { set ::nntpd::_NNTPD(ALLOW_POSTING) "y" } else { set ::nntpd::_NNTPD(ALLOW_POSTING) "n" } }
         "-disable_daemon" { if { [regexp -- {^[1yY]([eE][sS])?$} $val] } { set ::nntpd::_NNTPD(DISABLE_DAEMON) 1 } else { set ::nntpd::_NNTPD(DISABLE_DAEMON) 0 } }
         "-auth_proc" { set ::nntpd::_NNTPD(AUTH_PROC) $val }
         "-complain_email" { set ::nntpd::_NNTPD(COMPLAIN_EMAIL) $val }
         "-enable_post_notification" { if { [regexp -- {^[1yY]([eE][sS])?$} $val] } { set ::nntpd::_NNTPD(ENABLE_POST_NOTIFICATION) 1 } else { set ::nntpd::_NNTPD(ENABLE_POST_NOTIFICATION) 0 } }
         "-enable_post_forwarding" { if { [regexp -- {^[1yY]([eE][sS])?$} $val] } { set ::nntpd::_NNTPD(ENABLE_POST_FORWARDING) 1 } else { set ::nntpd::_NNTPD(ENABLE_POST_FORWARDING) 0 } }
         "-enable_mirroring" { if { [regexp -- {^[1yY]([eE][sS])?$} $val] } { set ::nntpd::_NNTPD(ENABLE_MIRRORING) 1 } else { set ::nntpd::_NNTPD(ENABLE_MIRRORING) 0 } }
         "-mirror_passwd" { set ::nntpd::_NNTPD(MIRROR_PASSWD) $val }
      }
   }

   #- Make sure the required options have been set
   if { [file isdirectory $::nntpd::_NNTPD(NEWS_ROOT)] == 0 } { puts "ERROR: Must specify -news_root option"; exit }
   
   #- Require packages and/or default some unspecified settings based on command line options
   if { $::nntpd::_NNTPD(ENABLE_POST_NOTIFICATION) || $::nntpd::_NNTPD(ENABLE_POST_FORWARDING) } \
   {
   	if { $::nntpd::_NNTPD(MAIL_HOST) == "" } { puts "ERROR: Email notification/forward was enabled but no -mailhost was defined."; exit }
   	package require ezsmtp 
   }
   if { $::nntpd::_NNTPD(ENABLE_MIRRORING) == 1 } \
   { 
	if { $::nntpd::_NNTPD(MIRROR_PASSWD) == "" } { puts "ERROR: Mirroring was enabled but no -mirror_passwd was defined."; exit }
   	package require nntp
   }
   if { $::nntpd::_NNTPD(ADMIN_EMAIL) == "" } { set ::nntpd::_NNTPD(ADMIN_EMAIL) "help@$::nntpd::_NNTPD(HOST)" }
   if { $::nntpd::_NNTPD(COMPLAIN_EMAIL) == "" } { set ::nntpd::_NNTPD(COMPLAIN_EMAIL) "abuse@$::nntpd::_NNTPD(HOST)" }
   if { $::nntpd::_NNTPD(DEBUG) == 1 } { set ::nntpd::_NNTPD(DISABLE_DAEMON) 1 }

   #-Check to see if news server is to require authentication
   if { $::nntpd::_NNTPD(DO_AUTH) } { set ::nntpd::_NNTPD(IS_AUTHORIZED) 0 } \
   else { set ::nntpd::_NNTPD(IS_AUTHORIZED) 1 }

#- end proc ::nntpd::init
}

proc ::nntpd::mirror_newsgroups { _NEWSGROUPS _SRC_HOST _DST_HOST _DST_MIRROR_PASSWD {_SRC_PORT 119} {_DST_PORT 119} } \
{
   package require nntp

   #- Open nntp connection to source and destination
   if { [catch {set SRC [::nntp::nntp $_SRC_HOST $_SRC_PORT]}] } { puts "Error: unable to connect to news source - $_SRC_HOST:$_SRC_PORT" } \
   elseif { [catch {set DST [::nntp::nntp $_DST_HOST $_DST_PORT]}] } { puts "Error: unable to connect to news destination - $_DST_HOST:$_DST_PORT" } \
   else \
   {
      #- Foreach newsgroup requested to be mirrored
      foreach _GRP $_NEWSGROUPS \
      {
         #- Ask destination for the last message it mirrored
         $DST mode_reader
         if { [catch {set resp [$DST group $_GRP]}] } { puts "Error: destination newsgroup, $_GRP, does not exist"; continue }
puts "In DST group: $resp"
         #- If it had gotten an article before, find the article's date
	 set LAST_ARTNUM ""
         if { [lindex $resp 0] != 0 } \
         { 
            set hdr [$DST head [lindex $resp 2]]
            foreach hdr_itm $hdr \
            {
	       if { [regexp -- {^X-Mirror: (.+)[ ]*$} $hdr_itm match temp] == 1 } \
	       { 
	          set LAST_ARTNUM [expr [lindex [split $temp :] 2] + 1]
	          break 
	       }
            }
         } \
         else \
         { 
	    set LAST_ARTNUM ""
         }

         #- Notify destination server that it should enter its mirroring mode to allow posting directly to server
         catch {$DST ihave $_DST_MIRROR_PASSWD}

	 #- Now talk to source
	 $SRC mode_reader
         if { [catch {set resp [$SRC group $_GRP]}] } { puts "Error: source newsgroup, $_GRP, does not exist"; continue }

	 #- resp's form: {total_messages first_message last_message group_name}
	 if { [lindex $resp 0] != 0 } \
	 {
	    set g_total [lindex $resp 0]
	    set g_first [lindex $resp 1]
	    set g_last [lindex $resp 2]
puts "In SRC group: $resp"
	    #- If no LAST_DATE found, then get all.  Else, ask only for the new ones.
	    set GET_END $g_last
	    if { $LAST_ARTNUM != "" } { set GET_START $LAST_ARTNUM } \
	    else { set GET_START $g_first }
	    if { $GET_START >= $g_last } \
	    {
	       #- No new articles
puts "No new articles to mirror"
	    } \
	    else \
	    {
puts "Have to mirror articles $GET_START - $GET_END"
	       #- Post the articles to DST starting with found
	       for {set j $GET_START} {$j <= $GET_END} {incr j} \
	       {
	          if { [catch {set art [$SRC article $j]}] == 0 } \
	   	  {
		     puts -nonewline "$j;"
		     flush stdout
                     #- Add header to indicate source of mirrored article
		     set art [linsert $art 0 "X-Mirror: $_SRC_HOST:$_SRC_PORT:$j"]
		     set art [linsert $art end ""]  ;#Hack of sorts 'cause the last line was getting dropped
		     $DST post [join "$art" \n]
	 	  }
	       }
	    }
	 } \
	 else { puts "ERROR: $resp" }
      #- end foreach newsgroup
      }
   #- end else
   }

   catch {$SRC quit}
   catch {$DST quit}
puts "done"
   return
#- end proc ::nntpd::mirror_newsgroups
}
